import os
from collections.abc import Sequence
from typing import TYPE_CHECKING, Optional

import dagster._check as check
from dagster._core.definitions.reconstruct import (
    load_def_in_module,
    load_def_in_package,
    load_def_in_python_file,
)

if TYPE_CHECKING:
    from dagster._core.workspace.autodiscovery import LoadableTarget

_DEFAULT_GRPC_TIMEOUT_IF_NO_ENV_VAR_SET = 60
_DEFAULT_REPOSITORY_TIMEOUT_IF_NO_ENV_VAR_SET = 180


def get_loadable_targets(
    *,
    python_file: Optional[str],
    module_name: Optional[str],
    package_name: Optional[str],
    autoload_defs_module_name: Optional[str],
    working_directory: Optional[str],
    attribute: Optional[str],
) -> Sequence["LoadableTarget"]:
    from dagster._core.workspace.autodiscovery import (
        LoadableTarget,
        autodefs_module_target,
        loadable_targets_from_python_file,
        loadable_targets_from_python_module,
        loadable_targets_from_python_package,
    )

    if python_file:
        return (
            [
                LoadableTarget(
                    attribute, load_def_in_python_file(python_file, attribute, working_directory)
                )
            ]
            if attribute
            else loadable_targets_from_python_file(python_file, working_directory)
        )
    elif module_name:
        return (
            [
                LoadableTarget(
                    attribute, load_def_in_module(module_name, attribute, working_directory)
                )
            ]
            if attribute
            else loadable_targets_from_python_module(module_name, working_directory)
        )
    elif package_name:
        return (
            [
                LoadableTarget(
                    attribute, load_def_in_package(package_name, attribute, working_directory)
                )
            ]
            if attribute
            else loadable_targets_from_python_package(package_name, working_directory)
        )
    elif autoload_defs_module_name:
        return [autodefs_module_target(autoload_defs_module_name, working_directory)]
    else:
        check.failed("invalid")


def max_rx_bytes() -> int:
    env_set = os.getenv("DAGSTER_GRPC_MAX_RX_BYTES")
    if env_set:
        return int(env_set)

    # default 100 MB
    return 100 * (10**6)


def max_send_bytes() -> int:
    env_set = os.getenv("DAGSTER_GRPC_MAX_SEND_BYTES")
    if env_set:
        return int(env_set)

    # default 100 MB
    return 100 * (10**6)


def default_grpc_timeout() -> int:
    env_set = os.getenv("DAGSTER_GRPC_TIMEOUT_SECONDS")
    if env_set:
        return int(env_set)

    return _DEFAULT_GRPC_TIMEOUT_IF_NO_ENV_VAR_SET


def default_repository_grpc_timeout() -> int:
    env_set = os.getenv("DAGSTER_REPOSITORY_GRPC_TIMEOUT_SECONDS")
    if env_set:
        return int(env_set)

    return max(_DEFAULT_REPOSITORY_TIMEOUT_IF_NO_ENV_VAR_SET, default_grpc_timeout())


def default_schedule_grpc_timeout() -> int:
    env_set = os.getenv("DAGSTER_SCHEDULE_GRPC_TIMEOUT_SECONDS")
    if env_set:
        return int(env_set)

    return default_grpc_timeout()


def default_sensor_grpc_timeout() -> int:
    env_set = os.getenv("DAGSTER_SENSOR_GRPC_TIMEOUT_SECONDS")
    if env_set:
        return int(env_set)

    return default_grpc_timeout()


def default_grpc_server_shutdown_grace_period():
    # Time to wait for calls to finish before shutting down the server
    # Defaults to the same as default_grpc_timeout() unless
    # DAGSTER_GRPC_SHUTDOWN_GRACE_PERIOD is set
    # default_repository_grpc_timeout(0) is omitted since that call is only
    # made during startup and is unlikely to need to be cleanly completed
    # in order to avoid downtime

    env_set = os.getenv("DAGSTER_GRPC_SHUTDOWN_GRACE_PERIOD")
    if env_set:
        return int(env_set)

    return max(
        default_grpc_timeout(), default_schedule_grpc_timeout(), default_sensor_grpc_timeout()
    )



import asyncio
import inspect
import tempfile
from abc import abstractmethod
from collections.abc import Awaitable
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Union
from uuid import uuid4

from typing_extensions import Self

from dagster._core.definitions.definitions_class import Definitions
from dagster._core.definitions.definitions_load_context import DefinitionsLoadContext
from dagster._core.errors import DagsterInvalidInvocationError
from dagster._core.storage.defs_state.base import DefsStateStorage
from dagster.components.component.component import Component
from dagster.components.core.context import ComponentLoadContext

if TYPE_CHECKING:
    from pydantic import BaseModel


class StateBackedComponent(Component):
    @classmethod
    def load(cls, attributes: Optional["BaseModel"], context: "ComponentLoadContext") -> Self:
        """Loads the component and marks its defs_state_key on the component tree."""
        loaded_component = super().load(attributes, context)
        context.component_tree.mark_component_defs_state_key(
            defs_state_key=loaded_component.get_defs_state_key(),
            component_path=context.component_path,
        )
        return loaded_component

    def get_defs_state_key(self) -> str:
        """Returns a key that uniquely identifies the state for this component."""
        return self.__class__.__name__

    def build_defs(self, context: ComponentLoadContext) -> Definitions:
        key = self.get_defs_state_key()

        with DefinitionsLoadContext.get().temp_state_path(key) as state_path:
            return self.build_defs_from_state(context, state_path=state_path)

    async def refresh_state(self) -> None:
        """Rebuilds the state for this component and persists it to the current StateStore."""
        key = self.get_defs_state_key()
        state_storage = DefsStateStorage.get()
        if state_storage is None:
            raise DagsterInvalidInvocationError(
                f"Attempted to refresh state of {key} without a StateStorage in context. "
                "This is likely the result of an internal framework error."
            )
        with tempfile.TemporaryDirectory() as temp_dir:
            state_path = Path(temp_dir) / key
            # handle either sync or async write_state_to_path functions
            if inspect.iscoroutinefunction(self.write_state_to_path):
                await self.write_state_to_path(state_path)
            else:
                await asyncio.to_thread(self.write_state_to_path, state_path)
            state_storage.upload_state_from_path(key, version=str(uuid4()), path=state_path)

    @abstractmethod
    def build_defs_from_state(
        self, context: ComponentLoadContext, state_path: Optional[Path]
    ) -> Definitions:
        """Given a state_path, builds a Definitions object based on the state
        contained at that path.

        Args:
            context (ComponentLoadContext): The context associated with this component load.
            state_path (Optional[Path]): The path to the state file. If no state has been
                previously written to the state store, this will be set to None.

        Returns:
            Definitions: The Definitions object built from the state.
        """

    @abstractmethod
    def write_state_to_path(self, state_path: Path) -> Union[None, Awaitable[None]]:
        """Fetches and writes required state to a local file. This method can be
        implemented as either sync or async.

        Args:
            state_path (Path): The path to the state file.
        """
        raise NotImplementedError()



from collections.abc import Mapping, Sequence
from enum import Enum
from typing import Union

import dagster._check as check
from dagster._config.config_type import ConfigTypeKind
from dagster._config.field_utils import EnvVar, IntEnvVar
from dagster._config.snap import ConfigFieldSnap, ConfigTypeSnap, minimal_config_for_type_snap
from dagster._config.stack import (
    EvaluationStackEntry,
    get_friendly_path_info,
    get_friendly_path_msg,
)
from dagster._config.traversal_context import ContextData
from dagster._config.type_printer import print_config_type_key_to_string
from dagster._record import IHaveNew, record, record_custom
from dagster._utils.error import SerializableErrorInfo


class DagsterEvaluationErrorReason(Enum):
    RUNTIME_TYPE_MISMATCH = "RUNTIME_TYPE_MISMATCH"
    MISSING_REQUIRED_FIELD = "MISSING_REQUIRED_FIELD"
    MISSING_REQUIRED_FIELDS = "MISSING_REQUIRED_FIELDS"
    FIELD_NOT_DEFINED = "FIELD_NOT_DEFINED"
    FIELDS_NOT_DEFINED = "FIELDS_NOT_DEFINED"
    SELECTOR_FIELD_ERROR = "SELECTOR_FIELD_ERROR"
    FAILED_POST_PROCESSING = "FAILED_POST_PROCESSING"
    FIELD_ALIAS_COLLISION = "FIELD_ALIAS_COLLISION"


@record
class FieldsNotDefinedErrorData:
    field_names: Sequence[str]


@record
class FieldAliasCollisionErrorData:
    field_name: str
    aliased_field_name: str


@record
class FieldNotDefinedErrorData:
    field_name: str


@record
class MissingFieldErrorData:
    field_name: str
    field_snap: ConfigFieldSnap


@record
class MissingFieldsErrorData:
    field_names: Sequence[str]
    field_snaps: Sequence[ConfigFieldSnap]


@record
class RuntimeMismatchErrorData:
    config_type_snap: ConfigTypeSnap
    value_rep: str


@record_custom
class SelectorTypeErrorData(IHaveNew):
    config_type_snap: ConfigTypeSnap
    incoming_fields: Sequence[str]

    def __new__(
        cls,
        *,
        config_type_snap: ConfigTypeSnap,
        incoming_fields: Sequence[str],
    ):
        check.inst_param(config_type_snap, "config_type_snap", ConfigTypeSnap)
        check.param_invariant(config_type_snap.kind == ConfigTypeKind.SELECTOR, "config_type")
        return super().__new__(
            cls,
            config_type_snap=config_type_snap,
            incoming_fields=incoming_fields,
        )


ERROR_DATA_UNION = Union[
    FieldNotDefinedErrorData,
    FieldsNotDefinedErrorData,
    MissingFieldErrorData,
    MissingFieldsErrorData,
    RuntimeMismatchErrorData,
    SelectorTypeErrorData,
    SerializableErrorInfo,
    FieldAliasCollisionErrorData,
]

ERROR_DATA_TYPES = ERROR_DATA_UNION.__args__  # type: ignore


@record
class EvaluationError:
    stack: EvaluationStackEntry
    reason: DagsterEvaluationErrorReason
    message: str
    error_data: ERROR_DATA_UNION


def _get_type_msg(type_in_context):
    if type_in_context.given_name is None:
        return ""
    else:
        return f' on type "{type_in_context.given_name}"'


def create_dict_type_mismatch_error(context: ContextData, config_value: object) -> EvaluationError:
    check.inst_param(context, "context", ContextData)

    path_msg, _path = get_friendly_path_info(context.stack)

    return EvaluationError(
        stack=context.stack,
        reason=DagsterEvaluationErrorReason.RUNTIME_TYPE_MISMATCH,
        message='Value {path_msg} must be dict. Expected: "{type_name}".'.format(
            path_msg=path_msg,
            type_name=print_config_type_key_to_string(
                context.config_schema_snapshot, context.config_type_key, with_lines=False
            ),
        ),
        error_data=RuntimeMismatchErrorData(
            config_type_snap=context.config_type_snap, value_rep=repr(config_value)
        ),
    )


def create_field_substitution_collision_error(
    context: ContextData, name: str, aliased_name: str
) -> EvaluationError:
    check.inst_param(context, "context", ContextData)
    check_config_type_in_context_has_fields(context, "context")
    return EvaluationError(
        stack=context.stack,
        reason=DagsterEvaluationErrorReason.FIELD_ALIAS_COLLISION,
        message=(
            f"Received both field '{name}' and field '{aliased_name}' in config. Please use one or"
            " the other."
        ),
        error_data=FieldAliasCollisionErrorData(field_name=name, aliased_field_name=aliased_name),
    )


def create_fields_not_defined_error(
    context: ContextData, undefined_fields: Sequence[str]
) -> EvaluationError:
    check.inst_param(context, "context", ContextData)
    check_config_type_in_context_has_fields(context, "context")
    check.list_param(undefined_fields, "undefined_fields", of_type=str)

    available_fields = sorted(context.config_type_snap.field_names)
    undefined_fields = sorted(undefined_fields)

    return EvaluationError(
        stack=context.stack,
        reason=DagsterEvaluationErrorReason.FIELDS_NOT_DEFINED,
        message=(
            f'Received unexpected config entries "{undefined_fields}" {get_friendly_path_msg(context.stack)}. '
            f'Expected: "{available_fields}."'
        ),
        error_data=FieldsNotDefinedErrorData(field_names=undefined_fields),
    )


def create_enum_type_mismatch_error(context: ContextData, config_value: object) -> EvaluationError:
    check.inst_param(context, "context", ContextData)

    return EvaluationError(
        stack=context.stack,
        reason=DagsterEvaluationErrorReason.RUNTIME_TYPE_MISMATCH,
        message=f"Value {get_friendly_path_msg(context.stack)} for enum type {context.config_type_snap.given_name} must be a string",
        error_data=RuntimeMismatchErrorData(
            config_type_snap=context.config_type_snap,
            value_rep=repr(config_value),
        ),
    )


def create_enum_value_missing_error(context: ContextData, config_value: object) -> EvaluationError:
    check.inst_param(context, "context", ContextData)

    return EvaluationError(
        stack=context.stack,
        reason=DagsterEvaluationErrorReason.RUNTIME_TYPE_MISMATCH,
        message=f"Value {get_friendly_path_msg(context.stack)} not in enum type {context.config_type_snap.given_name} got {config_value}",
        error_data=RuntimeMismatchErrorData(
            config_type_snap=context.config_type_snap,
            value_rep=repr(config_value),
        ),
    )


def check_config_type_in_context_has_fields(context: ContextData, param_name: str) -> None:
    check.param_invariant(ConfigTypeKind.has_fields(context.config_type_snap.kind), param_name)


def create_field_not_defined_error(context: ContextData, received_field: str) -> EvaluationError:
    check.inst_param(context, "context", ContextData)
    check_config_type_in_context_has_fields(context, "context")
    check.str_param(received_field, "received_field")

    return EvaluationError(
        stack=context.stack,
        reason=DagsterEvaluationErrorReason.FIELD_NOT_DEFINED,
        message=(
            'Received unexpected config entry "{received}" {path_msg}. Expected: "{type_name}".'.format(
                path_msg=get_friendly_path_msg(context.stack),
                type_name=print_config_type_key_to_string(
                    context.config_schema_snapshot, context.config_type_key, with_lines=False
                ),
                received=received_field,
            )
        ),
        error_data=FieldNotDefinedErrorData(field_name=received_field),
    )


def create_array_error(context: ContextData, config_value: object) -> EvaluationError:
    check.inst_param(context, "context", ContextData)
    check.param_invariant(context.config_type_snap.kind == ConfigTypeKind.ARRAY, "config_type")

    return EvaluationError(
        stack=context.stack,
        reason=DagsterEvaluationErrorReason.RUNTIME_TYPE_MISMATCH,
        message='Value {path_msg} must be list. Expected: "{type_name}"'.format(
            path_msg=get_friendly_path_msg(context.stack),
            type_name=print_config_type_key_to_string(
                context.config_schema_snapshot, context.config_type_key, with_lines=False
            ),
        ),
        error_data=RuntimeMismatchErrorData(
            config_type_snap=context.config_type_snap,
            value_rep=repr(config_value),
        ),
    )


def create_map_error(context: ContextData, config_value: object) -> EvaluationError:
    check.inst_param(context, "context", ContextData)
    check.param_invariant(context.config_type_snap.kind == ConfigTypeKind.MAP, "config_type")

    return EvaluationError(
        stack=context.stack,
        reason=DagsterEvaluationErrorReason.RUNTIME_TYPE_MISMATCH,
        message='Value {path_msg} must be dict. Expected: "{type_name}"'.format(
            path_msg=get_friendly_path_msg(context.stack),
            type_name=print_config_type_key_to_string(
                context.config_schema_snapshot, context.config_type_key, with_lines=False
            ),
        ),
        error_data=RuntimeMismatchErrorData(
            config_type_snap=context.config_type_snap,
            value_rep=repr(config_value),
        ),
    )


def create_missing_required_field_error(
    context: ContextData, expected_field: str
) -> EvaluationError:
    check.inst_param(context, "context", ContextData)
    check_config_type_in_context_has_fields(context, "context")

    missing_field_type = context.config_schema_snapshot.get_config_snap(
        context.config_type_snap.get_field(expected_field).type_key
    )

    return EvaluationError(
        stack=context.stack,
        reason=DagsterEvaluationErrorReason.MISSING_REQUIRED_FIELD,
        message=(
            'Missing required config entry "{expected}" {path_msg}. Sample config for missing'
            " entry: {minimal_config}"
        ).format(
            expected=expected_field,
            path_msg=get_friendly_path_msg(context.stack),
            minimal_config={
                expected_field: minimal_config_for_type_snap(
                    context.config_schema_snapshot, missing_field_type
                )
            },
        ),
        error_data=MissingFieldErrorData(
            field_name=expected_field, field_snap=context.config_type_snap.get_field(expected_field)
        ),
    )


def create_missing_required_fields_error(
    context: ContextData, missing_fields: Sequence[str]
) -> EvaluationError:
    check.inst_param(context, "context", ContextData)
    check_config_type_in_context_has_fields(context, "context")

    missing_fields = sorted(missing_fields)
    missing_field_snaps = list(map(context.config_type_snap.get_field, missing_fields))

    return EvaluationError(
        stack=context.stack,
        reason=DagsterEvaluationErrorReason.MISSING_REQUIRED_FIELDS,
        message=(
            "Missing required config entries {missing_fields} {path_msg}. Sample config for missing"
            " entries: {minimal_config}".format(
                missing_fields=missing_fields,
                path_msg=get_friendly_path_msg(context.stack),
                minimal_config={
                    field_snap.name: minimal_config_for_type_snap(
                        context.config_schema_snapshot,
                        context.config_schema_snapshot.get_config_snap(field_snap.type_key),
                    )
                    for field_snap in missing_field_snaps
                },
            )
        ),
        error_data=MissingFieldsErrorData(
            field_names=missing_fields, field_snaps=missing_field_snaps
        ),
    )


def create_scalar_error(context: ContextData, config_value: object) -> EvaluationError:
    check.inst_param(context, "context", ContextData)

    return EvaluationError(
        stack=context.stack,
        reason=DagsterEvaluationErrorReason.RUNTIME_TYPE_MISMATCH,
        message=(
            f'Invalid scalar {get_friendly_path_msg(context.stack)}. Value "{config_value}" of type '
            f'"{type(config_value)}" is not valid for expected type "{context.config_type_snap.given_name}".'
        ),
        error_data=RuntimeMismatchErrorData(
            config_type_snap=context.config_type_snap,
            value_rep=repr(config_value),
        ),
    )


def create_pydantic_env_var_error(
    context: ContextData, config_value: Union[EnvVar, IntEnvVar]
) -> EvaluationError:
    env_var_name = config_value.env_var_name

    correct_env_var = str({"env": env_var_name})
    return EvaluationError(
        stack=context.stack,
        reason=DagsterEvaluationErrorReason.RUNTIME_TYPE_MISMATCH,
        message=(
            f'Invalid use of environment variable wrapper. Value "{env_var_name}" is wrapped with'
            " EnvVar(), which is reserved for passing to structured pydantic config objects only."
            " To provide an environment variable to a run config dictionary, replace"
            f' EnvVar("{env_var_name}") with {correct_env_var}, or pass a structured RunConfig'
            " object."
        ),
        error_data=RuntimeMismatchErrorData(
            config_type_snap=context.config_type_snap,
            value_rep=repr(env_var_name),
        ),
    )


def create_selector_multiple_fields_error(
    context: ContextData, config_value: Mapping[str, object]
) -> EvaluationError:
    check.inst_param(context, "context", ContextData)

    defined_fields = sorted(context.config_type_snap.field_names)
    incoming_fields = sorted(list(config_value.keys()))

    return EvaluationError(
        stack=context.stack,
        reason=DagsterEvaluationErrorReason.SELECTOR_FIELD_ERROR,
        message=(
            f"You can only specify a single field {get_friendly_path_msg(context.stack)}. You specified {incoming_fields}. "
            f"The available fields are {defined_fields}"
        ),
        error_data=SelectorTypeErrorData(
            config_type_snap=context.config_type_snap, incoming_fields=incoming_fields
        ),
    )


def create_selector_multiple_fields_no_field_selected_error(
    context: ContextData,
) -> EvaluationError:
    check.inst_param(context, "context", ContextData)

    defined_fields = sorted(context.config_type_snap.field_names)

    return EvaluationError(
        stack=context.stack,
        reason=DagsterEvaluationErrorReason.SELECTOR_FIELD_ERROR,
        message=(
            f"Must specify a field {get_friendly_path_msg(context.stack)} if more than one field is defined. "
            f"Defined fields: {defined_fields}"
        ),
        error_data=SelectorTypeErrorData(
            config_type_snap=context.config_type_snap, incoming_fields=[]
        ),
    )


def create_selector_type_error(context: ContextData, config_value: object) -> EvaluationError:
    check.inst_param(context, "context", ContextData)

    return EvaluationError(
        stack=context.stack,
        reason=DagsterEvaluationErrorReason.RUNTIME_TYPE_MISMATCH,
        message=f"Value for selector type {get_friendly_path_msg(context.stack)} must be a dict",
        error_data=RuntimeMismatchErrorData(
            config_type_snap=context.config_type_snap, value_rep=repr(config_value)
        ),
    )


def create_selector_unspecified_value_error(context: ContextData) -> EvaluationError:
    check.inst_param(context, "context", ContextData)

    defined_fields = sorted(context.config_type_snap.field_names)

    return EvaluationError(
        stack=context.stack,
        reason=DagsterEvaluationErrorReason.SELECTOR_FIELD_ERROR,
        message=(
            f"Must specify the required field {get_friendly_path_msg(context.stack)}. Defined fields: {defined_fields}"
        ),
        error_data=SelectorTypeErrorData(
            config_type_snap=context.config_type_snap, incoming_fields=[]
        ),
    )


def create_none_not_allowed_error(context: ContextData) -> EvaluationError:
    check.inst_param(context, "context", ContextData)

    return EvaluationError(
        stack=context.stack,
        reason=DagsterEvaluationErrorReason.RUNTIME_TYPE_MISMATCH,
        message='Value {path_msg} must not be None. Expected "{type_name}"'.format(
            path_msg=get_friendly_path_msg(context.stack),
            type_name=print_config_type_key_to_string(
                context.config_schema_snapshot, context.config_type_key, with_lines=False
            ),
        ),
        error_data=RuntimeMismatchErrorData(
            config_type_snap=context.config_type_snap,
            value_rep=repr(None),
        ),
    )


def create_failed_post_processing_error(
    context: ContextData, original_value: object, error_data: SerializableErrorInfo
) -> EvaluationError:
    check.inst_param(context, "context", ContextData)
    check.inst_param(error_data, "error_data", SerializableErrorInfo)

    return EvaluationError(
        stack=context.stack,
        reason=DagsterEvaluationErrorReason.FAILED_POST_PROCESSING,
        message=(
            f"Post processing {get_friendly_path_msg(context.stack)} of original value {original_value} failed:\n{error_data.to_string()}"
        ),
        error_data=error_data,
    )


class PostProcessingError(Exception):
    """This is part of the formal API for implementing post_process
    methods on config types. Throw this error to indicate a
    that post processing cannot happen, and that the user
    must make a configuration and environment change in
    order resolve.
    """
